/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Algorithm.h - definicao da superclasse Algorithm, da qual devem descender
 *               todos os algoritmos implementados e introduzidos no GeoLab.
 *               Embora pareca o contrario, o objetivo da hierarquia de algo-
 *	         ritmos NAO e' restringir, e sim simplificar o trabalho com
 *	         o ambiente tornando mais facil a reutilizacao de codigo, 
 *		 permitindo uma organizacao logica mais eficiente quanto as
 *		 ferramentas de trabalho disponiveis e tambem ajudando na
 *               tarefa de se adequar o comportamento das ferramentas aos
 *		 'guidelines' do GeoLab.
 *
 *		 Do ponto de vista de reutilizacao externa, a introducao do
 *               conceito de hierarquia de algoritmos nao necessariamente
 *               desvirtua nosso objetivo inicial de proporcionar ferramentas
 *               independentes do ambiente. Na verdade, no ponto de vista do
 *               aplicador (i.e., aquele que usa o ambiente como bancada de
 *               desenvolvimento de ferramentas), tal conceito trata-se apenas 
 *               de um conjunto minimo de regras a serem atendidas para a 
 *               utilizacao do ambiente, e nao de "exportacao" de ferramentas,
 *               o que depende fortemente das necessidades especificas do
 *               aplicador.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    29.03.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *    07.01.93  13:10   C'esar N. Gon           Modos externos
 *
 */

#ifndef _ALGORITHM_DEFINED
#define _ALGORITHM_DEFINED

#include "global.h"
#include "utilities.h"
#include "basic.h"
#include "List.h"
#include "Obj.h"
#include "World.h"

/*****************************************************************************
  
  Para tentar diminuir o problema de nao termos a versao 3.0 do compilador
  C++, todos os metodos virtuais indefinidos em qualquer classe base sera'
  implementado de forma a produzir uma mensagem de erro do tipo:
  
  Warning: 'object type': 'method' not implemented
  
  Isso livra-me de ter que ficar contornando o problema da versao 2.0 nao
  permitir heranca sem redefinicao de metodos virtuais por mais de um ni-
  vel (somente uma classe base por hierarquia). Fique  claro  apenas que
  nestes casos, os metodos nao implementados realmente NAO tem  que  ser
  implementados, por tratar-se de uma classe abstrata.

******************************************************************************/
     
#ifndef _Obj_ptr_list
#define _Obj_ptr_list
declare(list,Obj_ptr)
#endif ~_Obj_ptr_list;

#ifndef _GraphicObj_ptr_list
#define _GraphicObj_ptr_list
declare(list,GraphicObj_ptr)
#endif ~_GraphicObj_ptr_list;

#ifndef _list_item_list
#define _list_item_list
declare(list,list_item)
#endif ~_list_item_list;

typedef list(Obj_ptr) ResultObjects;
typedef list(GraphicObj_ptr) GraphicObjects;
typedef list(list_item) SelectedObjects;

class Algorithm {
  public:

    virtual AlgorithmType type() {return ALG;}
    
    // is_able_to_handle() e' usada para determinar se o algoritmo pode ser
    // chamado para o conjunto de objetos selecionados no GeoLab. Isto e' util
    // para manter consistencia em relacao ao funcionamento dos algoritmos e
    // tambem para o GeoLab poder atualizar quais sao as opcoes disponiveis no
    // menu de algoritmos.
    
    virtual bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &) { 
        warning("Algorithm", "is_able_to_handle");
        return false; 
    }

    // animation() indica ao ambiente se o algoritmo possui ou nao qualquer
    // tipo de animacao (mesmo que em niveis inferiores). E' usada para tornar
    // mais facil a selecao de algoritmos quando a opcao de animacao esta'
    // selecionada.
    
    virtual bool animation() {
        return false;			// no warning, false is default
    }
    
    // funcao de ligacao entre a implementacao do algoritmo e o ambiente.
    // A tarefa de binder() resume-se a checar (opcionalmente) se o algoritmo
    // esta' apto a manipular a entrada que esta' lhe sendo fornecida, extrair
    // esta entrada em estruturas especificas e tambem realizar conversoes
    // entre diferentes representacoes caso seja necessario.
    
    virtual ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *, int animation_depth = 0, int reference_depth = 0) {
        ResultObjects res;
        warning("Algorithm", "binder");
        cout << "Animation Depth: " << animation_depth << "   "
             << "Reference Depth: " << reference_depth << endl;
        return res;    
    }

            
    // funcao de re-instanciacao de tipos criados externamente ao GeoLab
    // e introduzidos por meio de algoritmos geometricos ou de modos de
    // desenho externos. Deve retornar uma instancia do tipo t para que
    // o GeoLab seja capaz de realizar, por exemplo, carga do objeto
    // a partir de memoria secundaria.
    
    virtual Obj* instantiation(ObjType t) {
    	warning("Algorithm", "instantiation");
    	cout << "Object Type: " << t << endl;
    	return 0;
    }

    char *algorithm_name() {
	return _algorithm_name;
    }

    char *algorithm_name (char *n) {
	return _algorithm_name = n;
    }

    char *function_name() {
	return _function_name;
    }

    char *function_name (char *f) {
	return _function_name = f;
    }

    char *module_name() {
	return _module_name;
    }
    
    char *module_name (char *m) {
	return _module_name = m;
    }
    
  private:
    char *_algorithm_name;
    char *_function_name;
    char *_module_name;
};

class ExternalFunctionMode : public Algorithm {
  public:

    AlgorithmType type() {return EFM;}

    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &) { 
        return true; 
    }

    bool animation() {
        return true;			// no warning, true is default
    }
    
    // is_able_to_activate() e' usada pelo ambiente para determinar se
    // um dado modo externo pode ser ativado, de acordo com os objetos
    // selecionados no momento
    
    virtual bool is_able_to_activate(const SelectedObjects &, const GraphicObjects &, World *) {
        return true; 
    }

    virtual void installer(World *) {}
    virtual void uninstaller(World *) {}

    virtual void activator(World *) {}
    virtual void deactivator(World *) {}
    
    virtual icon_bits *icon() {return 0;}
    virtual icon_bits *cursor(int& xhot, int& yhot) {return (icon_bits *)(xhot = yhot = 0);}

    // objects_list_has_changed e' usada pelo ambiente p/ avisar ao modo
    // externo que houve mudan,ca no estado das listas objetos (SO e GO)
    // do GeoLab. Se este m'etodo retorna false, o modo externo e'
    // desativado, caso contr'ario, nada acontece (default).
    virtual bool objects_list_has_changed(const SelectedObjects &, const GraphicObjects &, World *) {return true;}
    
    virtual bool capture_events() {return true;}
    virtual bool click_select_down (World *, double, double) {return false;}
    virtual bool click_select_up (World *, double, double) {return false;}
    virtual bool click_adjust_down (World *, double, double) {return false;}
    virtual bool click_adjust_up (World *, double, double) {return false;}
    virtual bool double_click(World *, double, double) {return false;}
    virtual bool loc_move (World *, double, double) {return false;}
    virtual bool loc_drag (World *, double, double) {return false;}
    virtual bool ascii_key(World *, int) {return false;}
};


class ExternalObjectMode : public Algorithm {
  public:
    AlgorithmType type() {return EOM;}

    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &) { 
        return true; 
    }

    bool animation() {
        return true;			// no warning, true is default
    }
    
    // is_able_to_activate() e' usada pelo ambiente para determinar se
    // um dado modo externo pode ser ativado, de acordo com os objetos
    // selecionados no momento
    
    virtual bool is_able_to_activate(const SelectedObjects &, const GraphicObjects &, World *) {
	return true;
    }
    
    virtual void installer(World *) {}
    virtual void uninstaller(World *) {}

    virtual void activator(World *) {}
    virtual void deactivator(World *) {}
    
    virtual icon_bits *icon() {return 0;}
    virtual icon_bits *cursor(int& xhot, int& yhot) {return (icon_bits *)(xhot = yhot = 0);}

    // objects_list_has_changed e' usada pelo ambiente p/ avisar ao modo
    // externo que houve mudan,ca no estado das listas objetos (SO e GO)
    // do GeoLab. Se este m'etodo retorna false, o modo externo e'
    // desativado, caso contr'ario, nada acontece (default).
    virtual bool objects_list_has_changed(const SelectedObjects &, const GraphicObjects &, World *) {return true;}

    virtual bool capture_events() {return true;}
    virtual bool click_select_down (World *, double, double) {return false;}
    virtual bool click_select_up (World *, double, double) {return false;}
    virtual bool click_adjust_down (World *, double, double) {return false;}
    virtual bool click_adjust_up (World *, double, double) {return false;}
    virtual bool double_click(World *, double, double) {return false;}
    virtual bool loc_move (World *, double, double) {return false;}
    virtual bool loc_drag (World *, double, double) {return false;}
    virtual bool ascii_key(World *, int) {return false;}
};

// Uma vez que sao externos, os objetos componentes da hierarquia de algoritmos
// nao estao instanciados imediatamente apos a sua ligacao com o ambiente. 
// Assim sendo, faz-se necessaria uma funcao de ligacao (de fato, somente
// utilizada no momento do link externo) que informe ao GeoLab qual e' a
// instancia a ter seus metodos invocados nas tarefas subsequentes.
// Portanto, a cada algoritmo - digamos "alg" - deve haver uma funcao
// chamada "alg_link()" que retorna um pointer para uma instancia do
// algoritmo. Veja em External.h e nos arquivos de documentacao
// mais detalhes sobre o esquema de ligacao externa.

#endif ~_ALGORITHM_DEFINED;





