/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Diameter.C - algoritmos para o calculo do diametro de um conjunto de
 *              pontos, de poligonos convexos, em forma de estrela e simples.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.05.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Diameter.h"
#include "Misc.h"
#include "ConvexHull.h"
#include "Real.h"

// includes para animacao

#include "Circle.h"

static GraphicSegment2D *anim_s;
static GraphicCircle *anim_c;

#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_circle	((Circle *)anim_c->obj)

void diameter_animation_init(World *w) {
  static bool FIRST = true;
  
  if (FIRST) {
    anim_s = new GraphicSegment2D();
    anim_c = new GraphicCircle();
    FIRST = false;
  }
  anim_s->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
  anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
}
 
/*
 * Antipodal Pairs - implementation of the algorithm proposed in Shamos and
 *                   Preparata's book - "Computational Geometry"
 *                   Springer-Verlag, 1985, pg. 174.
 *
 */
 
/*
 * To determinate if a given simple polygon is ordered counterclockwise, we
 * use the winding number algorithm.
 *
 */
 
static bool simple_polygon_ordered_counterclockwise(list(Point2D_ptr) &points) {
    if (points.size() > 2) {
      int wn, colinear = 0;
      list_item a, b, c;
      Point2D ref;
      SideType s, t;
      
      a = points.first(); b = points.succ(a); c = points.succ(b);
      ref = gravitycentre(*points[a], *points[b], *points[c]);
      wn = winding_number(points, ref);
      
      if (wn == 0) {
      	if (points[a]->left_turn(*points[b], *points[c]))
      	  s = Right;
      	else
      	  s = Left;
      	  
      	do {
       	  a = points.cyclic_succ(a); b = points.cyclic_succ(a); c = points.cyclic_succ(b);
       	  t = points[a]->which_side(*points[b], *points[c]);
       	  if (t == Colinear)
       	    colinear++;
        } while (t != s && colinear < points.size());
    
        if (colinear == points.size())
          return true;
          
        ref = gravitycentre(*points[a], *points[b], *points[c]);
        wn = winding_number(points, ref);
      }
      
      return (wn < 0);
    }
    else
      return true;
}
 
list(Segment2D_ptr) antipodal_pairs(list(Point2D_ptr) points, World *w, int animation_depth, int reference_depth) {
    list(Segment2D_ptr) result;
    list_item p, q, q0;
    bool stop = false;

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      diameter_animation_init(w);
    ANIMATION_END(w, animation_depth, reference_depth)
        
    if (points.size() > 3) {
      // 1st - assert that the polygon is ordered counterclockwise
      if (!simple_polygon_ordered_counterclockwise(points)) {
        list(Point2D_ptr) aux;
        Point2D_ptr p;
      
        forall(p, points) {
          aux.push(p);
        }
        points.clear();
        forall(p, aux) {
          points.append(p);
        }
      }
    
      p = points.last();
      q = points.cyclic_succ(p);

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(points[p]->X(), points[p]->Y(), 6);
        anim_c->draw(w, XorMode);
        anim_circle->set(points[q]->X(), points[q]->Y(), 6);
        anim_c->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      while (points[points.cyclic_succ(q)]->area(*points[p], *points[points.cyclic_succ(p)]) >
             points[q]->area(*points[p], *points[points.cyclic_succ(p)])) {

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          frame_stop_mark(w);
          anim_circle->set(points[q]->X(), points[q]->Y(), 6);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
            
        q = points.cyclic_succ(q);
  
        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(points[q]->X(), points[q]->Y(), 6);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
            
      }
   
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_circle->set(points[points.cyclic_succ(p)]->X(), points[points.cyclic_succ(p)]->Y(), 10);
        anim_c->draw(w, XorMode);
        anim_circle->set(points[q]->X(), points[q]->Y(), 10);
        anim_c->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      q0 = q;
      while(q != points.first()) {

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(points[p]->X(), points[p]->Y(), 6);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)

        p = points.cyclic_succ(p);
        result.push(new Segment2D(*points[p], *points[q]));
        
        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(points[p]->X(), points[p]->Y(), 6);
          anim_c->draw(w, XorMode);
          anim_segment->set(*points[p], *points[q]);
          anim_s->draw(w, XorMode);
          frame_stop_mark(w);  
        ANIMATION_END(w, animation_depth, reference_depth)
        
        while (points[points.cyclic_succ(q)]->area(*points[p], *points[points.cyclic_succ(p)]) >
               points[q]->area(*points[p], *points[points.cyclic_succ(p)])) {

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(points[q]->X(), points[q]->Y(), 6);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)

          q = points.cyclic_succ(q);

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(points[q]->X(), points[q]->Y(), 6);
            anim_c->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth, reference_depth)

          if (p != q0 || q != points.first()) {

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_segment->set(*points[p], *points[q]);
              anim_s->draw(w, XorMode);
              frame_stop_mark(w);  
            ANIMATION_END(w, animation_depth, reference_depth)
        
            result.push(new Segment2D(*points[p], *points[q]));
          }
          else {
            stop = true;
            break;
          }
        }
        if (!stop && points[points.cyclic_succ(q)]->area(*points[p], *points[points.cyclic_succ(p)]) ==
                     points[q]->area(*points[p], *points[points.cyclic_succ(p)]))
          if (p != q0 || q != points.last()) {

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_segment->set(*points[p], *points[q]);
              anim_s->draw(w, XorMode);
              frame_stop_mark(w);  
            ANIMATION_END(w, animation_depth, reference_depth)
        
            result.push(new Segment2D(*points[p], *points[points.cyclic_succ(q)]));
          }
      }
    }
    
    return result;        
}

Segment2D_ptr diameter_convex_polygon(Polygon_ptr pol, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) points;

    points = pol->copy_vertices(); 
    if (points.size() > 3) {
      list(Segment2D_ptr) anti_pairs;
      Segment2D_ptr seg;
      Segment2D res;
      double max, calc;
      anti_pairs = antipodal_pairs(points, w, animation_depth, reference_depth);
      max = -1;
      forall(seg, anti_pairs) {
        calc = seg->P1().e_distance_sqr(seg->P2());
        if (calc > max) {
    	  max = calc;
    	  res = *seg;
        }
        delete seg;
      }
      
      return (Segment2D *)res.reproduct();
    }
    else
      if (points.size() == 1)
        return new Segment2D(*points.head(), *points.head());
      else
        if (points.size() == 2)
          return new Segment2D(*points.head(), *points[points.succ(points.first())]);
        else {
          Point2D_ptr a, b, c;
          double dab, dbc, dca;
          a = points.head();
          b = points[points.succ(points.first())];
          c = points[points.succ(points.succ(points.first()))];
          dab = a->e_distance_sqr(*b);
          dbc = b->e_distance_sqr(*c);
          dca = c->e_distance_sqr(*a);
          if (dab >= dbc && dab >= dca)
            return new Segment2D(*a, *b);
          else
            if (dbc >= dab && dbc >= dca)
              return new Segment2D(*b, *c);
            else
              return new Segment2D(*c, *a);
        }
}

Segment2D_ptr diameter_star_shaped_polygon(Polygon_ptr pol, World *w, int animation_depth, int reference_depth) {
    Segment2D_ptr res;
    
    Polygon *ch = new Polygon(ch_starshaped_polygon(pol->copy_vertices(), w, animation_depth+1, reference_depth));
    res = diameter_convex_polygon(ch, w, animation_depth, reference_depth);
    delete ch;
    return res;
}

Segment2D_ptr diameter_simple_polygon(Polygon_ptr pol, World *w, int animation_depth, int reference_depth) {
    Segment2D_ptr res;
    
    Polygon *ch = new Polygon(ch_simple_polygon(pol->copy_vertices(), w, animation_depth+1, reference_depth));
    res = diameter_convex_polygon(ch, w, animation_depth, reference_depth);
    delete ch;
    return res;
}

Segment2D_ptr diameter_generic_polygon(Polygon_ptr pol, World *w, int animation_depth, int reference_depth) {
    Segment2D_ptr res;
    
    Polygon *ch = new Polygon(ch_generic_polygon(pol->copy_vertices(), w, animation_depth+1, reference_depth));
    res = diameter_convex_polygon(ch, w, animation_depth, reference_depth);
    delete ch;
    return res;
}

Segment2D_ptr diameter_set_of_points(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    Segment2D_ptr res;
    
    Polygon *ch = new Polygon(graham_scan(points, w, animation_depth+1, reference_depth));
    res = diameter_convex_polygon(ch, w, animation_depth, reference_depth);
    delete ch;
    return res;
}

Segment2D_ptr diameter_set_of_points_bf(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list_item it_p, it_farthest_p;
    Segment2D_ptr res = 0;		// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      diameter_animation_init(w);
    ANIMATION_END(w, animation_depth, reference_depth)
        
    if (points.size() > 1) {
      Point2D_ptr p1 = 0, p2 = 0;
      double maior_aux, maior = -1;
      forall_list_items(it_p, points) {
        forall_list_items(it_farthest_p, points) {

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment->set(*points[it_p], *points[it_farthest_p]);
            anim_s->draw(w, XorMode);
            frame_stop_mark(w);
            anim_s->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)

          if (it_p != it_farthest_p && (maior_aux = points[it_p]->e_distance_sqr(*points[it_farthest_p])) > maior) {
          
            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              if (p1) {
              	anim_segment->set(*p1, *p2);
              	anim_s->draw(w, XorMode);
              }
            ANIMATION_END(w, animation_depth, reference_depth)

            p1 = points[it_p]; p2 = points[it_farthest_p];

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_segment->set(*p1, *p2);
              anim_s->draw(w, XorMode);
              frame_stop_mark(w);
            ANIMATION_END(w, animation_depth, reference_depth)
            
            maior = maior_aux;
          }
        }
      }
      res = new Segment2D(*p1, *p2);
    }
    
    return res;
}

// Modelagem

static list(Point2D_ptr) extract_points(const SelectedObjects &SO, const GraphicObjects &GO) {
    list(Point2D_ptr) res;
    list_item it;
    forall(it, SO) {			 	// is_able_to_handle deve ser
      res.push((Point2D *)GO[it]->obj);		// chamado antes para certificar
    }						// que sao todos pontos
    return res;
}

bool SetOfPoints::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    
    forall(it, SO) {
      if (GO[it]->type() != Point2D_t) {
      	SO.init_iterator();
      	return false;
      }
    }
    
    return true;
}

bool InputIsConvexPolygon::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t && GO[SO.head()]->subtype() == Convex_t)
      return true;

    return false;
}

ResultObjects AntipodalPairs::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Segment2D_ptr) anti_pairs;
    list(Point2D_ptr) points;
    Segment2D_ptr seg;
    Point2D_ptr point;
        
    if (!SO.empty()) {
      points = ((Polygon *)GO[SO.head()]->obj)->copy_vertices();
      anti_pairs = antipodal_pairs(points, w, animation_depth, reference_depth); 
      forall(seg, anti_pairs) {
        result.push((Obj_ptr)seg);
      }
 
      forall(point, points) {
        delete point;
      }
    }
      
    return result;
}

ResultObjects DiameterConvexPolygon::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_convex_polygon(((Polygon *)GO[SO.head()]->obj), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

bool DiameterStarShapedPolygon::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t && (GO[SO.head()]->subtype() == StarShaped_t || GO[SO.head()]->subtype() == Convex_t))
      return true;

    return false;
}

ResultObjects DiameterStarShapedPolygon::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_star_shaped_polygon(((Polygon *)GO[SO.head()]->obj), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

bool DiameterSimplePolygon::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t && (GO[SO.head()]->subtype() == Simple_t || GO[SO.head()]->subtype() == StarShaped_t || GO[SO.head()]->subtype() == Convex_t))
      return true;

    return false;
}

ResultObjects DiameterSimplePolygon::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_simple_polygon(((Polygon *)GO[SO.head()]->obj), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

bool DiameterGenericPolygon::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t)
      return true;

    return false;
}

ResultObjects DiameterGenericPolygon::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_generic_polygon(((Polygon *)GO[SO.head()]->obj), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

ResultObjects DiameterSetOfPoints::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_set_of_points(extract_points(SO, GO), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

ResultObjects DiameterSetOfPointsBF::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Segment2D_ptr seg;
        
    if (!SO.empty()) {
      seg = diameter_set_of_points_bf(extract_points(SO, GO), w, animation_depth, reference_depth);
      if (seg) {
      	Real *r;
        result.push((Obj_ptr)seg);
        r = new Real(seg->P1().e_distance(seg->P2()));
        r->obj_volatile(TRUE);
        result.push((Obj_ptr)r);
      }
    }
      
    return result;
}

Algorithm* antipodal_pairs_link() {
    return new AntipodalPairs();
}

Algorithm* diameter_convex_polygon_link() {
    return new DiameterConvexPolygon();
}

Algorithm* diameter_star_shaped_polygon_link() {
    return new DiameterStarShapedPolygon();
}

Algorithm* diameter_simple_polygon_link() {
    return new DiameterSimplePolygon();
}

Algorithm* diameter_generic_polygon_link() {
    return new DiameterGenericPolygon();
}

Algorithm* diameter_set_of_points_link() {
    return new DiameterSetOfPoints();
}

Algorithm* diameter_set_of_points_bf_link() {
    return new DiameterSetOfPointsBF();
}




