/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Kernel.C - implementacao do algoritmo para o calculo do kernel (nucleo) de
 *            um poligono simples.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    24.04.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Kernel.h"
#include "Misc.h"
#include "Segment2D.h"
#include "Line2D.h"
#include "Ray2D.h"

// includes para animacao

#include "Circle.h"

// variaveis para animacao

static GraphicCircle *anim_c;
static GraphicSegment2D *anim_s;
static GraphicRay2D *anim_r;

#define anim_circle	((Circle *)anim_c->obj)
#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_ray	((Ray2D *)anim_r->obj)

void kernel_animation_init(World *w) {
    static bool FIRST = true;
    
    if (FIRST) {
      anim_c = new GraphicCircle();
      anim_s = new GraphicSegment2D();
      anim_r = new GraphicRay2D();
      FIRST = false;
    }
    anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_s->set_all(w, FOREGROUND, BACKGROUND, 0, 1, false);
    anim_r->set_all(w, FOREGROUND, BACKGROUND, 0, 1, false);
}

void draw_kernel(World *w, const list(Point2D_ptr) &K, bool open) {
    GraphicSegment2D s;
    list_item it;
    
    s.set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    forall_list_items(it, K) {
      if (!open || (it != K.last() && open)) {
        ((Segment2D *)s.obj)->set(*K[it], *K[K.cyclic_succ(it)]);
        s.draw(w, XorMode);
      }
    }
}

void draw_support_lines(World *w, Point2D *vi, Point2D *Fi, Point2D *Li) {
    if (!Fi->at_infinity()) {
      anim_circle->set(Fi->X(), Fi->Y(), 6);
      anim_c->draw(w, XorMode);
      anim_ray->set(*vi, *Fi);
      anim_r->draw(w, XorMode);
    }
    if (!Li->at_infinity()) {
      anim_circle->set(Li->X(), Li->Y(), 6);
      anim_c->draw(w, XorMode);
      anim_ray->set(*vi, *Li);
      anim_r->draw(w, XorMode);
    }
}

/*
 * To determinate if a given simple polygon is ordered counterclockwise, we
 * use the winding number algorithm.
 *
 */
 
static bool simple_polygon_ordered_counterclockwise(list(Point2D_ptr) &points) {
    if (points.size() > 2) {
      int wn, colinear = 0;
      list_item a, b, c;
      Point2D ref;
      SideType s, t;
      
      a = points.first(); b = points.succ(a); c = points.succ(b);
      ref = gravitycentre(*points[a], *points[b], *points[c]);
      wn = winding_number(points, ref);
      
      if (wn == 0) {
      	if (points[a]->left_turn(*points[b], *points[c]))
      	  s = Right;
      	else
      	  s = Left;
      	  
      	do {
       	  a = points.cyclic_succ(a); b = points.cyclic_succ(a); c = points.cyclic_succ(b);
       	  t = points[a]->which_side(*points[b], *points[c]);
       	  if (t == Colinear)
       	    colinear++;
        } while (t != s && colinear < points.size());
    
        if (colinear == points.size())
          return true;
          
        ref = gravitycentre(*points[a], *points[b], *points[c]);
        wn = winding_number(points, ref);
      }
      
      return (wn < 0);
    }
    else
      return true;
}
 
/*
 * Kernel of a Simple Polygon in O(n) - implementation of the algorithm
 *                                      proposed by Lee and Preparata
 *
 * (reference: D. T. Lee and F. P. Preparata - "An optimal algorithm for 
 *             finding the kernel of a polygon" - J. ACM, 
 *             v.26 pg. 415-421, 1979)
 *
 */

bool lies_to_right(Point2D *PFLi, Point2D *PFLant, Point2D *v1, Point2D *v2) {
    if (PFLi->at_infinity()) {
      Line2D lv1v2(*v1, *v2);
      Line2D lantFLi(*PFLant, *PFLi);

      return lv1v2.cross_to_right(lantFLi);
    }

    Point2D pv1, pv2;
    pv1.setXY(v1->X(), v1->Y());
    pv2.setXY(v2->X(), v2->Y());    

    return PFLi->right_turn(pv1, pv2);
}
 
bool lies_to_left(Point2D *PFLi, Point2D *PFLant, Point2D *v1, Point2D *v2) {
    if (PFLi->at_infinity()) {
      Line2D lv1v2(*v1, *v2);
      Line2D lantFLi(*PFLant, *PFLi);
      
      return lv1v2.cross_to_left(lantFLi);
    }
    
    Point2D pv1, pv2;
    pv1.setXY(v1->X(), v1->Y());
    pv2.setXY(v2->X(), v2->Y());    

    return PFLi->left_turn(pv1, pv2);
}

void free_kernel(list(Point2D_ptr) &K, bool open, World *w, int animation_depth, int reference_depth) {
    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      draw_kernel(w, K, open);
    ANIMATION_END(w, animation_depth, reference_depth)
              	
    Point2D_ptr p;
    forall(p, K) {
      delete p;
    }
    K.clear();
}

// Nota Importante:
//
//  O algoritmo proposto por Lee & Preparata faz uso de "pontos no infinito"
// durante a construcao do Kernel. Isto tras grandes problemas para a imple-
// mentacao, visto que a manipulacao de pontos no infinito nao e', de forma
// alguma, trivial. As primitivas 'intersect' e 'intersection' tiveram de
// ser reformuladas (Ray2D e Line2D) para suportar perguntas ou calculos do
// tipo 'intersect(a, b)' onde a e b sao pontos, possivelmente no infinito.
//  Outro serio problema que ocorreu e' que ao atualizar o Kernel do polygono
// em um dado passo do algoritmo, era necessario verificar se pontos no
// infinito nao estavam sendo afetados, pois sua representacao depende
// (no caso deste algoritmo) de um outro vertice (que nao esta' no
// infinito) como ponto de partida. A eliminacao deste vertice de partida
// torna possivelmente invalido o valor do ponto no infinito. Para solucionar
// este problema, quando um novo vertice e' inserido no kernel, e' verificado
// (em tempo constante, ja' que a localidade de referencia do problema e'
// constante) se nao estamos violando a propriedade do ponto no infinito e,
// caso estejamos, o valor e' corrigido em funcao do novo ponto de partida.
 
list(Point2D_ptr) kernel_simple_polygon(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) K;

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      kernel_animation_init(w);
    ANIMATION_END(w, animation_depth, reference_depth)

    remove_colinear_points(points);
    
    if (points.size() < 4) {
      Point2D_ptr p;
      forall(p, points) {
      	K.append(p);
      }
    }
    else {
      // 1st - assert that the polygon is ordered counterclockwise
      if (!simple_polygon_ordered_counterclockwise(points)) {
        list(Point2D_ptr) aux;
        Point2D_ptr p;
      
        forall(p, points) {
          aux.push(p);
        }
        points.clear();
        forall(p, aux) {
          points.append(p);
        }
      }
    
      // 2nd - v0 must be a reflex vertex
      list_item v0 = points.first();
      bool found = false;
      do {
        if (points[points.cyclic_pred(v0)]->right_turn(*points[v0], *points[points.cyclic_succ(v0)])) {
          found = true;
          break;
        }
        v0 = points.cyclic_succ(v0);
      } while (v0 != points.first());
      
      // 3rd - if NO reflex vertex then the polygon is convex ... nothing to do!
      
      if (!found) {
        Point2D_ptr p;
        forall(p, points) {
          K.append(p);
        }
      }
      else {
        // now the algorithm
        
        list_item prev, vi, next;
        list_item Li, Fi, search;
        list_item wt, wt1, ws, ws1, wr, wr1, wu, wu1;
        list_item wl_pos, wll_pos;
        Point2D *dummyLi, *dummyFi, *wl, *wll;
        bool open = true, wll_found, error;
        
        // initial state of K
        
        dummyLi = new Point2D();
        dummyFi = new Point2D();
        
        dummyLi->set_at_infinity((2 * points[v0]->X()) - points[points.cyclic_pred(v0)]->X(), 
                                 (2 * points[v0]->Y()) - points[points.cyclic_pred(v0)]->Y());
        dummyFi->set_at_infinity((2 * points[v0]->X()) - points[points.cyclic_succ(v0)]->X(),
                                 (2 * points[v0]->Y()) - points[points.cyclic_succ(v0)]->Y());
                                 
        // set initial values of Fi and Li

        Li = K.push(dummyLi);
        K.push((Point2D *)points[v0]->reproduct());
        Fi = K.push(dummyFi);
        
        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(points[v0]->X(), points[v0]->Y(), 10);
          anim_c->draw(w, XorMode);
          draw_kernel(w, K, open);
          frame_stop_mark(w);
          anim_circle->set(points[v0]->X(), points[v0]->Y(), 10);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
          
        // main loop
        
        vi = points.cyclic_succ(v0);
        prev = v0;
        next = points.cyclic_succ(vi);
        
        while (vi != points.cyclic_pred(v0)) {
        	
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(points[vi]->X(), points[vi]->Y(), 10);
            anim_c->draw(w, XorMode);
            draw_support_lines(w, points[vi], K[Fi], K[Li]);
            frame_stop_mark(w);
            draw_support_lines(w, points[vi], K[Fi], K[Li]);
          ANIMATION_END(w, animation_depth, reference_depth)
          
       	  if (points[prev]->right_turn(*points[vi], *points[next])) {
       	    // vertex is reflex
            if (lies_to_right(K[Fi], K[K.cyclic_succ(Fi)], points[vi], points[next])) {
       	      
       	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	        anim_ray->set(*points[next], *points[vi]);
       	        anim_r->draw(w, XorMode);
       	        frame_stop_mark(w);
       	      ANIMATION_END(w, animation_depth, reference_depth)
       	    
              search = Fi;
              do {
                wt1 = search;
                wt = K.cyclic_succ(wt1);
                
                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_segment->set(*K[wt], *K[wt1]);
                  anim_s->draw(w, XorMode);
                  frame_stop_mark(w);
                  anim_s->draw(w, XorMode);
                ANIMATION_END(w, animation_depth, reference_depth)
                
                if (Ray2D(*points[next], *points[vi]).intersect(*K[wt], *K[wt1]))
                  break;
                else
                  search = K.cyclic_succ(search);
              } while (search != Li);     

              if (search == Li) {
              	// empty kernel
                free_kernel(K, open, w, animation_depth, reference_depth);
              	return K;
              }
              else {
              	wl = new Point2D(Ray2D(*points[next], *points[vi]).intersection(*K[wt], *K[wt1]));
              	
              	ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_circle->set(wl->X(), wl->Y(), 4);
                  anim_c->draw(w, XorMode);
                  frame_stop_mark(w);
              	ANIMATION_END(w, animation_depth, reference_depth)
              	
              	search = Fi; 
              	wll_found = false;
              	while (!open || (K.pred(search) && open)) {
              	  ws = search;
              	  ws1 = K.cyclic_pred(ws);

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
                    anim_segment->set(*K[ws], *K[ws1]);
                    anim_s->draw(w, XorMode);
                    frame_stop_mark(w);
                    anim_s->draw(w, XorMode);
                  ANIMATION_END(w, animation_depth, reference_depth)
                
              	  if (Ray2D(*points[next], *points[vi]).intersect(*K[ws], *K[ws1])) {
              	    wll_found = true;
              	    wll = new Point2D(Ray2D(*points[next], *points[vi]).intersection(*K[ws], *K[ws1]));

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      frame_stop_mark(w);
              	    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	    break;
              	  }
              	  else
              	    if (open)
              	      search = K.pred(search);
              	    else
              	      search = K.cyclic_pred(search);
              	      
          	  if (search == Fi) {
                    // empty kernel (maybe an error)
                    free_kernel(K, open, w, animation_depth, reference_depth);
              	    return K;
              	  }
              	}
              	if (wll_found) {

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
                    anim_circle->set(wl->X(), wl->Y(), 4);
                    anim_c->draw(w, XorMode);
                    anim_circle->set(wll->X(), wll->Y(), 4);
                    anim_c->draw(w, XorMode);
              	    draw_kernel(w, K, open);
              	  ANIMATION_END(w, animation_depth, reference_depth)

              	  wl_pos = K.insert(wl, wt1, after);
              	  wll_pos = K.insert(wll, ws, before);
              	  
              	  // atualiza o valor do ponto no infinito, se necessario
              	  if (K[wt]->at_infinity())
              	    K[wt]->set_at_infinity((2 * wl->X()) - K[wt1]->X(),
              	    			   (2 * wl->Y()) - K[wt1]->Y());
              	    
                  if (K[ws1]->at_infinity())
                    K[ws1]->set_at_infinity((2 * wll->X()) - K[ws]->X(),
                    			    (2 * wll->Y()) - K[ws]->Y());

              	  list_item it = K.cyclic_pred(wl_pos);
              	  do {
              	    list_item aux = K.cyclic_pred(it);
              	    delete K[it];
              	    K.del_item(it);
              	    it = aux;
              	  } while (it != wll_pos);

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
              	    draw_kernel(w, K, open);
              	    frame_stop_mark(w);
              	  ANIMATION_END(w, animation_depth, reference_depth)
              	
              	}
              	else {
              	  double acur = points[next]->angle(*points[vi]);
              	  double aini = K[K.succ(K.first())]->angle(*K.head());
              	  double afin = K[K.pred(K.last())]->angle(*K.tail());

 		  if (aini < afin)
 		    if (acur > PI)
  		      aini += 2 * PI;
  		    else
  		      afin -= 2 * PI;
		    
              	  if (acur <= aini && acur >= afin) {

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wl->X(), wl->Y(), 4);
                      anim_c->draw(w, XorMode);
                      draw_kernel(w, K, open);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
               	    list_item it = K.first();
               	    wl_pos = K.insert(wl, wt1, after);
               	    do {
              	      list_item aux = K.cyclic_succ(it);
              	      delete K[it];
              	      K.del_item(it);
              	      it = aux;
              	    } while (it != wl_pos);

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
              	      draw_kernel(w, K, open);
              	      frame_stop_mark(w);
              	    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	  }
              	  else {
              	    search = K.last();
              	    error = true;
              	    do {
              	      wr = search;
              	      wr1 = K.pred(wr);

                      ANIMATION_BEGIN(w, animation_depth, reference_depth)
                        anim_segment->set(*K[wr1], *K[wr]);
                        anim_s->draw(w, XorMode);
                        frame_stop_mark(w);
                        anim_s->draw(w, XorMode);
                      ANIMATION_END(w, animation_depth, reference_depth)
                
                      if (Ray2D(*points[next], *points[vi]).intersect(*K[wr1], *K[wr])) {
                      	error = false;
                        break;
                      }
                      else
                        search = K.pred(search);              	      
              	    } while (search != K.first());

                    if (error) {
                      // empty kernel (or error, due to floating point inaccuracies)
                      free_kernel(K, open, w, animation_depth, reference_depth);
                      return K;
                    }              	  
                    
              	    wll_found = true;
              	    wll = new Point2D(Ray2D(*points[next], *points[vi]).intersection(*K[wr1], *K[wr]));
              	    
                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      frame_stop_mark(w);
                      anim_circle->set(wl->X(), wl->Y(), 4);
                      anim_c->draw(w, XorMode);
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      draw_kernel(w, K, open);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
                    wl_pos = K.insert(wl, wt1, after);
              	    wll_pos = K.insert(wll, wr, before);
              	  
              	    open = false;  
              	    list_item it = K.cyclic_pred(wl_pos);
              	    do {
              	      list_item aux = K.cyclic_pred(it);
              	      delete K[it];
              	      K.del_item(it);
              	      it = aux;
              	    } while (it != wll_pos);

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
              	      draw_kernel(w, K, open);
              	      frame_stop_mark(w);
              	    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	  }
              	}
              } 

              if (wll_found)
                Fi = wll_pos;
              else {

                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  draw_kernel(w, K, open);
                ANIMATION_END(w, animation_depth, reference_depth)
              	
              	dummyFi = new Point2D();
              	if (!(*points[vi] == *wl))
                  dummyFi->set_at_infinity((2 * wl->X()) - points[vi]->X(),
                                           (2 * wl->Y()) - points[vi]->Y());
                else
                  dummyFi->set_at_infinity((2 * points[vi]->X()) - points[next]->X(),
                                           (2 * points[vi]->Y()) - points[next]->Y());
                
                Fi = K.push(dummyFi);

                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  draw_kernel(w, K, open);
                  frame_stop_mark(w);
                ANIMATION_END(w, animation_depth, reference_depth)
              	
              }

       	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	        anim_ray->set(*points[next], *points[vi]);
       	        anim_r->draw(w, XorMode);
       	      ANIMATION_END(w, animation_depth, reference_depth)
       	    
            }
            else {
              search = Fi;
              error = true;
              do {
              	wt = search;
              	wt1 = K.cyclic_succ(wt);
              	if (lies_to_right(K[wt1], K[wt], points[next], K[wt])) {
              	  Fi = wt;
              	  error = false;
              	  break;
              	}
              	else
              	  search = K.cyclic_succ(search);
              } while (search != Fi);

              if (error) {
                // empty kernel (or error, due to floating point inaccuracies)
                free_kernel(K, open, w, animation_depth, reference_depth);
                return K;
              }              	  
            }
            
            search = Li;
            while (!open || (K.succ(search) && open)) {
              wu = search;
              wu1 = K.cyclic_succ(wu);
              if (lies_to_left(K[wu1], K[wu], points[next], K[wu])) {
                Li = wu;
                break;
              }
              else
                search = K.cyclic_succ(search);
                
              if (search == Li) {
                // empty kernel (maybe an error)
                free_kernel(K, open, w, animation_depth, reference_depth);
                return K;
              }
            }
          }
       	  else {
       	    // vertex is convex
            if (lies_to_right(K[Li], K[K.cyclic_pred(Li)], points[vi], points[next])) {
              
       	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	        anim_ray->set(*points[vi], *points[next]);
       	        anim_r->draw(w, XorMode);
       	        frame_stop_mark(w);
       	      ANIMATION_END(w, animation_depth, reference_depth)
       	    
              search = Li;
              do {
                wt = search;
                wt1 = K.cyclic_pred(wt);

                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_segment->set(*K[wt1], *K[wt]);
                  anim_s->draw(w, XorMode);
                  frame_stop_mark(w);
                  anim_s->draw(w, XorMode);
                ANIMATION_END(w, animation_depth, reference_depth)
                
                if (Ray2D(*points[vi], *points[next]).intersect(*K[wt1], *K[wt]))
                  break;
                else
                  search = K.cyclic_pred(search);
              } while (search != Fi);     

              if (search == Fi) {
              	// empty kernel
                free_kernel(K, open, w, animation_depth, reference_depth);
              	return K;
              }
              else {
              	wl = new Point2D(Ray2D(*points[vi], *points[next]).intersection(*K[wt1], *K[wt]));

              	ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  anim_circle->set(wl->X(), wl->Y(), 4);
                  anim_c->draw(w, XorMode);
                  frame_stop_mark(w);
              	ANIMATION_END(w, animation_depth, reference_depth)
              	
              	search = Li; 
              	wll_found = false;
              	while (!open || (K.succ(search) && open)) {
              	  ws1 = search;
              	  ws = K.cyclic_succ(ws1);

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
                    anim_segment->set(*K[ws1], *K[ws]);
                    anim_s->draw(w, XorMode);
                    frame_stop_mark(w);
                    anim_s->draw(w, XorMode);
                  ANIMATION_END(w, animation_depth, reference_depth)
                
              	  if (Ray2D(*points[vi], *points[next]).intersect(*K[ws1], *K[ws])) {
              	    wll_found = true;
              	    wll = new Point2D(Ray2D(*points[vi], *points[next]).intersection(*K[ws1], *K[ws]));

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      frame_stop_mark(w);
              	    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	    break;
              	  }
              	  else
              	    if (open)
              	      search = K.succ(search);
              	    else
              	      search = K.cyclic_succ(search);
              	    
              	  if (search == Li) {
                    // empty kernel (maybe an error)
                    free_kernel(K, open, w, animation_depth, reference_depth);
                    return K;
              	  }
              	}
              	if (wll_found) {

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
                    anim_circle->set(wl->X(), wl->Y(), 4);
                    anim_c->draw(w, XorMode);
                    anim_circle->set(wll->X(), wll->Y(), 4);
                    anim_c->draw(w, XorMode);
                    draw_kernel(w, K, open);
                  ANIMATION_END(w, animation_depth, reference_depth)

              	  wl_pos = K.insert(wl, wt1, after);
              	  wll_pos = K.insert(wll, ws, before);

              	  // atualiza o valor do ponto no infinito, se necessario
              	  if (K[wt1]->at_infinity())
              	    K[wt1]->set_at_infinity((2 * wl->X()) - K[wt]->X(),
              	    		 	    (2 * wl->Y()) - K[wt]->Y());
              	    
                  if (K[ws]->at_infinity())
                    K[ws]->set_at_infinity((2 * wll->X()) - K[ws1]->X(),
                    		           (2 * wll->Y()) - K[ws1]->Y());

              	  list_item it = K.cyclic_succ(wl_pos);
              	  do {
              	    list_item aux = K.cyclic_succ(it);
              	    delete K[it];
              	    K.del_item(it);
              	    it = aux;
              	  } while (it != wll_pos);

                  ANIMATION_BEGIN(w, animation_depth, reference_depth)
                    draw_kernel(w, K, open);
                    frame_stop_mark(w);
                  ANIMATION_END(w, animation_depth, reference_depth)
              	
              	}
              	else {
              	  double acur = points[vi]->angle(*points[next]);
              	  double aini = K[K.succ(K.first())]->angle(*K.head());
              	  double afin = K[K.pred(K.last())]->angle(*K.tail());

 		  if (aini < afin)
 		    if (acur > PI)
  		      aini += 2 * PI;
  		    else
  		      afin -= 2 * PI;
		    
              	  if (acur <= aini && acur >= afin) {

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wl->X(), wl->Y(), 4);
                      anim_c->draw(w, XorMode);
                      draw_kernel(w, K, open);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
               	    list_item it = K.last();
               	    wl_pos = K.insert(wl, wt1, after);
               	    do {
              	      list_item aux = K.cyclic_pred(it);
              	      delete K[it];
              	      K.del_item(it);
              	      it = aux;
              	    } while (it != wl_pos);

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      draw_kernel(w, K, open);
                      frame_stop_mark(w);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	  }
              	  else {
              	    search = K.first();
              	    error = true;
              	    do {
              	      wr = search;
              	      wr1 = K.succ(wr);

                      ANIMATION_BEGIN(w, animation_depth, reference_depth)
                        anim_segment->set(*K[wr1], *K[wr]);
                        anim_s->draw(w, XorMode);
                        frame_stop_mark(w);
                        anim_s->draw(w, XorMode);
                      ANIMATION_END(w, animation_depth, reference_depth)
                
                      if (Ray2D(*points[vi], *points[next]).intersect(*K[wr1], *K[wr])) {
                      	error = false;
                        break;
                      }
                      else
                        search = K.succ(search);              	      
              	    } while (search != K.last());
                    
                    if (error) {
                      // empty kernel (or error, due to floating point inaccuracies)
                      free_kernel(K, open, w, animation_depth, reference_depth);
                      return K;
                    }              	  
                    
              	    wll_found = true;
              	    wll = new Point2D(Ray2D(*points[vi], *points[next]).intersection(*K[wr1], *K[wr]));
              	    
                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      frame_stop_mark(w);
                      anim_circle->set(wl->X(), wl->Y(), 4);
                      anim_c->draw(w, XorMode);
                      anim_circle->set(wll->X(), wll->Y(), 4);
                      anim_c->draw(w, XorMode);
                      draw_kernel(w, K, open);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
                    wl_pos = K.insert(wl, wt1, after);
              	    wll_pos = K.insert(wll, wr, after);
              	  
                    open = false;
              	    list_item it = K.cyclic_succ(wl_pos);
              	    do {
              	      list_item aux = K.cyclic_succ(it);
              	      delete K[it];
              	      K.del_item(it);
              	      it = aux;
              	    } while (it != wll_pos);

                    ANIMATION_BEGIN(w, animation_depth, reference_depth)
                      draw_kernel(w, K, open);
                      frame_stop_mark(w);
                    ANIMATION_END(w, animation_depth, reference_depth)
              	
              	  }
              	}
              } 
              
              if (Segment2D(*points[vi], *wl).point_in(*points[next])) {
                search = Fi;
                error = true;
                do {
                  wt = search;
              	  wt1 = K.cyclic_succ(wt);
              	  if (lies_to_right(K[wt1], K[wt], points[next], K[wt])) {
              	    Fi = wt;
              	    error = false;
              	    break;
              	  }
              	  else
              	    search = K.cyclic_succ(search);
                } while (search != Fi);

                if (error) {
                  // empty kernel (or error, due to floating point inaccuracies)
                  free_kernel(K, open, w, animation_depth, reference_depth);
                  return K;
                }              	  
              }
              else
                Fi = wl_pos;
              	
              if (wll_found)
              	if (Segment2D(*points[vi], *wll).point_in(*points[next]))
              	  Li = wll_pos;
              	else {
                  search = wl_pos;
                  while (!open || (K.succ(search) && open)) {
                    wu = search;
                    wu1 = K.cyclic_succ(wu);
                    if (lies_to_left(K[wu1], K[wu], points[next], K[wu])) {
                      Li = wu;
                      break;
                    }
                    else
                      search = K.cyclic_succ(search);
                    
                    if (search == wl_pos) {
                      // empty kernel (maybe an error)
                      free_kernel(K, open, w, animation_depth, reference_depth);
                      return K;
                    }
                  }
              	}
              else {
                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  draw_kernel(w, K, open);
                ANIMATION_END(w, animation_depth, reference_depth)
              	
              	dummyLi = new Point2D();
              	if (!(*points[vi] == *wl))
                  dummyLi->set_at_infinity((2 * wl->X()) - points[vi]->X(),
                                           (2 * wl->Y()) - points[vi]->Y());
                else
                  dummyLi->set_at_infinity((2 * points[next]->X()) - points[vi]->X(),
                                           (2 * points[next]->Y()) - points[vi]->Y());
                
                Li = K.append(dummyLi);

                ANIMATION_BEGIN(w, animation_depth, reference_depth)
                  draw_kernel(w, K, open);
                ANIMATION_END(w, animation_depth, reference_depth)
              	
              }

       	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	        anim_ray->set(*points[vi], *points[next]);
       	        anim_r->draw(w, XorMode);
       	      ANIMATION_END(w, animation_depth, reference_depth)
       	    
            }
            else {
              search = Fi;
              error = true;
              do {
                wt = search;
              	wt1 = K.cyclic_succ(wt);
              	if (lies_to_right(K[wt1], K[wt], points[next], K[wt])) {
              	  Fi = wt;
              	  error = false;
              	  break;
              	}
              	else
            	  search = K.cyclic_succ(search);
              } while (search != Fi);

              if (error) {
                // empty kernel (or error, due to floating point inaccuracies)
                free_kernel(K, open, w, animation_depth, reference_depth);
                return K;
              }              	  
                
              if (!open) {
                search = Li;
                while (!open || (K.succ(search) && open)) {
                  wu = search;
                  wu1 = K.cyclic_succ(wu);
                  if (lies_to_left(K[wu1], K[wu], points[next], K[wu])) {
                    Li = wu;
                    break;
                  }
                  else
                    search = K.cyclic_succ(search);
                    
                  if (search == Li) {
                    // empty kernel (maybe an error)
                    free_kernel(K, open, w, animation_depth, reference_depth);
                    return K;
                  }
                }
              }
            }
       	  }
       	  
          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_circle->set(points[vi]->X(), points[vi]->Y(), 10);
            anim_c->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)
          
          prev = vi;
       	  vi = points.cyclic_succ(vi);
       	  next = points.cyclic_succ(vi);
       	}
      }
    }
    
    return K;
}

bool Kernel::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t/* && GO[SO.head()]->subtype() != Generic_t*/)
      return true;
    
    return false;
}

ResultObjects Kernel::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    
    if (!SO.empty()) {
      Polygon *p = ((Polygon *)GO[SO.head()]->obj);
      list(Point2D_ptr) kernel = kernel_simple_polygon(p->copy_vertices(), w, animation_depth, reference_depth);
      if (kernel.size() != 0) {
      	Polygon *res = new Polygon(kernel);
      	res->subtype(Convex_t); res->closed(true);
        result.push(res);
      }
    }
    
    return result;
}

Algorithm* kernel_simple_polygon_link() {
    return new Kernel();
}

