/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * dynamic.C - modulo de tratamento de eventos no modo de movimentacao dinamica
 *             de objetos (chamadas sucessivas ao ultimo algoritmo invocado).
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.12.91  08:45 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "dynamic.h"
#include "color.h"
#include "line_settings.h"
#include "algorithms.h"
#include "objects.h"
#include "trapsignal.h"
#include "Obj.h"
#include "Point2D.h"
#include "compose.h"

#define NO_IF_GLUED true

bool dynamic_toggle_events(World *w, Xv_Window xwin, Event *event) {
    static bool resize_ev = false;
    static bool in_operation = false;
    static bool released_ctrl = false;
        
    switch (event_action(event)) {
      case ACTION_SELECT:
      case ACTION_ADJUST:
      case ACTION_MENU:
        if (event_is_down(event))
          in_operation = true;
        else {
          in_operation = false;
          if (released_ctrl) {
            resize_ev = false;
            released_ctrl = false;
          }
        }
        break;
        
      case SHIFT_LEFTCTRL:
        if (event_is_down(event) && !in_operation)
            resize_ev = true;
        else
          if (event_is_up(event) && resize_ev && in_operation)
            released_ctrl = true;
          else
            if (event_is_up(event) && resize_ev && !in_operation)
              resize_ev = false;
        break;
    }

    bool consumed;    
    if (resize_ev)
      consumed = resize_single_events(w, xwin, event);
    else
      consumed = move_copy_single_events(w, xwin, event);
    if (!consumed)
      return dynamic_events(w, xwin, event);
    else
      return false;
}

bool move_copy_single_events(World *world, Xv_Window, Event *event) {
    static GraphicObj *current_obj = 0;
    static double xi = 0, yi = 0;
    double X = world->virtual_xcoord(event_x(event));
    double Y = world->virtual_ycoord(event_y(event));
    double dx, dy;

    dx = X - xi; xi = X;
    dy = Y - yi; yi = Y;
        
    switch (event_action(event)) {
      case ACTION_SELECT:
      case ACTION_ADJUST:
        if (event_is_down(event)) {
          if (!current_obj) {
            if (current_obj = objects_object_selected_at(X, Y, NO_IF_GLUED))
              if (!current_obj->selected())
                current_obj = 0;
              else {
                if (event_action(event) == ACTION_ADJUST) {
                  GraphicObj *dupl = current_obj->reproduct();
                if (dupl)
                  objects_insert_and_select(dupl);
              	}
              	unsigned long fg, bg;
              	current_obj->get_color(fg, bg);
              	current_obj->set_color(world, get_pixel_from_value(WHITE), get_pixel_from_value(WHITE));
              	current_obj->draw(world, CopyMode);
              	objects_repaint_all_but(current_obj, world);
              	current_obj->unselect();
              	current_obj->set_color(world, fg, bg);
              	current_obj->draw(world, XorMode);
              	return false;
              }
          }
        }
        else 
          if (current_obj) {
            current_obj->draw(world, XorMode);
            current_obj->translate(dx, dy);
            current_obj->select();
            current_obj->draw(world, CopyMode);
            current_obj = 0;
            return false;
          }
        break;
        
      case LOC_DRAG:
        if (current_obj) {
          current_obj->draw(world, XorMode);
          current_obj->translate(dx, dy);
          current_obj->draw(world, XorMode);
          return false;
        }
        break;
    }
    
    return true;
}

bool resize_single_events(World *world, Xv_Window, Event *event) {
    static GraphicObj *current_obj = 0;
    static Int_Mode mode;
    double X = world->virtual_xcoord(event_x(event));
    double Y = world->virtual_ycoord(event_y(event));

    if (event_action(event) == ACTION_SELECT)
      mode = LeftButton;
    if (event_action(event) == ACTION_ADJUST)
      mode = MiddleButton;
      
    switch (event_action(event)) {
      case ACTION_SELECT:
      case ACTION_ADJUST:
        if (event_is_down(event)) {
          if (!current_obj) {
            if (current_obj = objects_object_selected_by_handle_at(X, Y, NO_IF_GLUED))
              if (!current_obj->selected())
                current_obj = 0;
              else {
                unsigned long fg, bg;
                current_obj->get_color(fg, bg);
                current_obj->set_color(world, get_pixel_from_value(WHITE), get_pixel_from_value(WHITE));
                current_obj->draw(world, CopyMode);
                objects_repaint_all_but(current_obj, world);
                current_obj->unselect();
                current_obj->set_color(world, fg, bg);
                current_obj->resize_start(world, X, Y, mode);
                current_obj->draw(world, XorMode);
                return false;
              }
          } 
        }
        else 
          if (current_obj) {
            current_obj->draw(world, XorMode);
            current_obj->resize_end(world, X, Y, mode);
            current_obj->select();
            current_obj->draw(world, CopyMode);
            current_obj = 0;
            return false;
          }
        break;
        
      case LOC_DRAG:
        if (current_obj) {
          current_obj->draw(world, XorMode);
          current_obj->resize_change(world, X, Y, mode);
          current_obj->draw(world, XorMode);
          return false;
        }
        break;
    }
    
    return true;
}

void release_objects(ResultObjects &res) {
    Obj *obj;
    
    forall(obj, res) {
      delete obj;
    }	
    res.clear();
}

void draw_objects(ResultObjects res, World *w, GC gc) {
    Obj *obj;
    
    forall(obj, res) {
      obj->draw(w, gc);
      if (obj->type() == Point2D_t)
        ((Point2D *)obj)->draw_circular_handles(w, gc, false, true);
    }	
}

bool dynamic_events(World *w, Xv_Window, Event *event) {
    static ResultObjects res;
    static bool first = true, dynamic = false;
    static GC gc;
        
    switch(event_action(event)) {
      case ACTION_SELECT:
      case ACTION_ADJUST:
        if (event_is_down(event)) {
          if (!LAST_PLAY && (dynamic = last_algorithm_is_able_to_handle_current_objects())) {
            if (first) {
              gc = w->CreateGC(0, NULL);
              first = false;
            }
            w->SetForeground(gc, FOREGROUND);
            w->SetBackground(gc, BACKGROUND);
            w->SetLineAttributes(gc, LINE_WIDTH, LINE_STYLE, 0, 0);
            w->SetFunction(gc, GXxor);
            res = last_algorithm_called(w);
            draw_objects(res, w, gc);
            set_external_signal_handler();
          }
        }
        else {
          if (dynamic) {
            set_internal_signal_handler();
            draw_objects(res, w, gc);
            release_objects(res);
            if (DISCARD_INPUT)
              objects_delete_selected(w);
            else
              if (UNSELECT_INPUT) {
                objects_unselect(w);
                objects_repaint(w);
              }
            if (!DYNAMIC_DISCARD_OUTPUT) {
              res = last_algorithm_called(w);
              objects_install_created_objects(w, OUTPUT_SELECTED, res, 
                                              last_algorithm_called_name(),
                                              last_algorithm_called_module_name());
            }
            else
              objects_repaint(w);
          }
          dynamic = false;
        }
        break;
        
      case LOC_DRAG:
        if (dynamic) {
          draw_objects(res, w, gc);
          release_objects(res);
          res = last_algorithm_called(w);
          draw_objects(res, w, gc);
        }
        break;
    }
    
    return false;
}







