/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * GWB.h - contem a definicao da classe GWB, descrita por Mantyla
 *               em "An Introduction to Solid Modeling"
 *
 *               link flag (-lm)
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    25.10.91  11:30	Welson R. Jacometti	Arquivo Criado
 *
 */

#ifndef _GWB_DEFINED
#define _GWB_DEFINED

#include <iostream.h>
#include <math.h>
#include "basic.h"
#include "List.h"
#include "global.h"
#include "Coords3D.h"
#include "tools.h"

// basic typedef's

typedef long Id;

// orientations

enum Orientation { PLUS, MINUS };


// epslon for geometric tests

const double EPS =              1e-6; // tolerance for geometric tests
const double BIGEPS =           1e-4; // a more permissive tolerance



// Load & save...

enum EulerOperator { MVFS , KVFS , MEV , SMEV , KEV , MEF , SMEF , KEF , KEMR , MEKR , SMEKR , KFMRH , MFKRH , RINGMV };

class EulerOperatorRepresentation
{
  public:
    Id f1,f2,v1,v2,v3,v4;
    double x,y,z;
    EulerOperator name;
};

typedef EulerOperatorRepresentation* EulerOperatorRepresentation_ptr;

declare (list,EulerOperatorRepresentation_ptr);

typedef list(EulerOperatorRepresentation_ptr) OperatorSequence;
typedef list(EulerOperatorRepresentation_ptr) *OperatorSequence_ptr;



class EulerSequence
{
  public:
    EulerSequence() {;}
    ~EulerSequence(){ kill(); }
    void kill();
    void push(Id f1,Id f2, Id v1, Id v2, Id v3, Id v4, double x, double y, double z,EulerOperator name);
    void append(Id f1,Id f2, Id v1, Id v2, Id v3, Id v4, double x, double y, double z,EulerOperator name);
    void pop(Id& f1,Id& f2, Id& v1, Id& v2, Id& v3, Id& v4, double& x, double& y, double& z,EulerOperator& name);
    void show();
    
    OperatorSequence it;
};

typedef EulerSequence* EulerSequence_ptr;
     

// forward declarations

class Solid;
class Face;
class Loop;
class Vertex;
class Edge;
class HalfEdge;


// GWB structure's root class

class GWB {
public:

  // management methods

  GWB() { firsts = NULL; }
  ~GWB() { killAll(); }

  bool killSolid( Id sno );
  void killSolid( Solid *s );
  void killAll();

  // query methods

  Id getNextSolidId();
  Solid *getSolid ( Id sno ) const;
  HalfEdge *getHalfEdge( Id sno, Id fno, Id vno );
  HalfEdge *getHalfEdge( Id sno, Id fno, Id v1no, Id v2no );  
  Vertex *getVertex( Id sno, Id vno );
  Solid *getFirsts() const { return firsts; }

  // hi-level Euler's operators methods
  
  bool mvfs(Solid *&s, Id sno, Id fn, Id vn, float x, float y, float z = 0);
  bool mvfs(Id sno, Id fn, Id vn, float x, float y, float z = 0) {
    Solid *s;
    return mvfs(s, sno, fn, vn, x, y, z);
  }
  bool mvfs(Id fn, Id vn, float x, float y, float z = 0) {
    Solid *s;
    return mvfs(s, getNextSolidId(), fn, vn, x, y, z);
  }
  bool kvfs( Id sno );

  bool mev( Id sno, Id f1, Id f2, Id v1, Id v2, Id v3, Id v4,
               float x, float y, float z = 0 );
  bool smev( Id sno, Id f1, Id v1, Id v4, float x, float y, float z = 0 );
  bool kev( Id sno, Id f, Id v1, Id v2 );

  bool mef( Id sno, Id f1, Id v1, Id v2, Id v3, Id v4, Id f2 );
  bool smef( Id sno, Id f1, Id v1, Id v3, Id f2 );
  bool kef( Id sno, Id f, Id v1, Id v2 );

  bool kemr( Id sno, Id f, Id v1, Id v2 );
  bool mekr( Id sno, Id f, Id v1, Id v2, Id v3, Id v4 );
  bool smekr( Id sno, Id f, Id v1, Id v3 );

  // load & save ...
  
  EulerSequence_ptr invert(Id solid);
  void build(Id solid_id,const EulerSequence& sequence);
  
  // duplicate ...

  void duplicate(GWB& gwb);


  // geometric support methods
  Solid *circle(float cx, float cy, float rad, float h, int n);
  Solid *torus(float r1, float r2, int nf1, int nf2);

  Solid *firsts;               // head of the list of all solids

};


// GWB component classes

class Solid {
public:
  Solid() {sgwb = 0;sfaces = 0;sedges = 0;sverts = 0;nexts = 0;prevs = 0;}
  Solid(GWB *gwb);
  ~Solid();

  void addface(Face *face);
  void delface(Face *face);

  void addedge(Edge *edge);
  void deledge(Edge *edge);

  void addvertex(Vertex *vertex);
  void delvertex(Vertex *vertex);

  Face *findFace(Id fno);
  Vertex *findVertex(Id vno);
  void checkv();
  void getmaxnames(Id &maxf, Id &maxv);
    
  void arc(Id f, Id v, float cx, float cy, float rad, float h, float phi1, float phi2, int n);

  void solidls(int sw);
  void listsolid();

  bool mev( Id f1, Id f2, Id v1, Id v2, Id v3, Id v4,
               float x, float y, float z = 0 );
  bool smev(Id fn, Id vn1, Id vn2, float x, float y, float z = 0);
  bool kev(Id fnr, Id v1, Id v2);

  bool mef( Id f1, Id v1, Id v2, Id v3, Id v4, Id f2 );
  bool smef(Id vert1, Id vert2, Id fac1, Id fac2);
  bool kef(Id v1, Id v2, Id fno);
   
  bool kemr( Id f, Id v1, Id v2 );
  bool mekr( Id f, Id v1, Id v2, Id v3, Id v4 );
  bool smekr( Id f, Id v1, Id v3 );

  void revert();
  void rsweep(int nfaces);

void ringmv(Id v1, Id v2, Id f1, Id f2, int out);
  void aringmv(int f1, int f2);

  double svolume();

  // Replication ...

  Solid* duplicate();

  GWB    *sgwb;         // pointer to gwb ("owner")
  Id     solidno;       // solid identifier
  Face   *sfaces;	// pointer to list of faces
  Edge   *sedges;	// pointer to list of edges
  Vertex *sverts;	// pointer to list of vertices
  Solid  *nexts;        // pointer to next solid
  Solid  *prevs;        // pointer to previous solid
};

class Face {
public:
  Face() { fsolid = 0;flout = 0;floops = 0;nextf = 0;prevf = 0;info = 0;}
  Face(Solid *s) { s->addface(this); floops = 0; flout = 0; }
  
  void addloop(Loop *this_loop);
  void delloop(Loop *this_loop);

  HalfEdge *findhe1(Id vn);
  HalfEdge *findhe2(Id vn1, Id vn2);

  void facels(int sw);
  void loopglue();

  Id          faceno;	// face identifier
  Solid       *fsolid;	// back pointer to solid
  Loop        *flout;	// pointer to outer loop
  Loop        *floops;	// pointer to list of loops
  Face        *nextf;	// pointer to next face
  Face        *prevf;	// pointer to previous face
  SpaceCoords feq;	// face equation
  void        *info;
};

typedef Face* Face_ptr;

class Loop {
public:
  Loop() {ledg = 0;lface = 0;nextl = 0;prevl = 0;info = 0;}
  Loop(Face *f) { f->addloop(this); }
  
  bool faceeq(SpaceCoords eq);
  double loopar();

  HalfEdge *ledg;	// pointer to ring of halfedges
  Face     *lface;	// back pointer to face
  Loop     *nextl;	// pointer to next loop
  Loop     *prevl;	// pointer to previous loop
  void     *info;
};

typedef Loop* Loop_ptr;

class Edge {
public:
  Edge() {he1 = 0;he2 = 0;nexte = 0;preve = 0;info = 0;}
  Edge(Solid *s) { s->addedge(this); }

  HalfEdge *he1;	// pointer to right halfedge
  HalfEdge *he2;	// pointer to left halfedge
  Edge     *nexte;	// pointer to next edge
  Edge     *preve;	// pointer to previous edge
  void     *info;        // specific to aplications
};

typedef Edge*  Edge_ptr;

class Vertex
{
  public:
    Vertex() { vedge = 0; nextv = 0; prevv = 0; info = 0;}
    Vertex(Solid *s) { s->addvertex(this); vedge = 0; }
    
    void listneighbors();

    // Query methods...
      
    double X() { return vcoord[Space_XCoord];}
    double Y() { return vcoord[Space_YCoord];}
    double Z() { return vcoord[Space_ZCoord];}
    
    Id           vno; // vertex identifier
    HalfEdge    *vedge; // pointer to an adjacent halfedge
    Vertex      *nextv;	// pointer to next vertex
    Vertex      *prevv;	// pointer to previous vertex
    SpaceCoords vcoord;	// vertex coordinates
    void        *info;
};

typedef Vertex* Vertex_ptr;

class HalfEdge {
public:

    HalfEdge() { edg = 0;vtx = 0;wloop = 0;nxt = 0;prv = 0;info = 0;}
 // Query Methods...
   
  double X() { return vtx->X();}
  double Y() { return vtx->Y();}
  double Z() { return vtx->Z();}

 
  Edge     *edg;	// pointer to parent edge
  Vertex   *vtx;	// pointer to starting vertex
  Loop     *wloop;	// back pointer to loop
  HalfEdge *nxt;	// pointer to next halfedge
  HalfEdge *prv;	// pointer to previous halfedge
  void     *info;
};

typedef HalfEdge*  HalfEdge_ptr;

// euler operators (those appliable to a particular class are within them)

/*
HalfEdge *mate( HalfEdge *he ) {
    return (((he) == (he)->edg->he1) ? (he)->edg->he2 : (he)->edg->he1);
}
*/

#define mate(he) (((he) == (he)->edg->he1) ? (he)->edg->he2 : (he)->edg->he1)


// Error handler functions

EXPORT void SetWarningHandler( void (*Handler) (const char *) );
EXPORT void SetFatalHandler( void (*Handler) (const char *, int) );



// Acochambracao...

EXPORT void lmev(HalfEdge *he1, HalfEdge *he2, Id vn, float x, float y, float z);

EXPORT Face *lmef(HalfEdge *he1, HalfEdge *he2, Id nf);

EXPORT Face_ptr kill_rays(Vertex_ptr v);

#endif _GWB_DEFINED;













